#!/usr/bin/env python
##
##  edit_mp3.py - tiny mp3 frame editor
##  by Yusuke Shinyama
##
##  usage: $ ./edit_mp3.py '0-100,1000-2000' a.mp3 b.mp3 > c.mp3
##

import sys, re, fileinput
from struct import unpack
stderr = sys.stderr

PAT = re.compile(r'^([0-9]*)-([0-9]*)$')
def range2list(s, n0, n1):
  r = []
  for i in s.split(','):
    i = i.strip()
    if i.isdigit():
      n = int(i)
      if n0 <= n and n <= n1:
        r.append(n)
      else:
        raise ValueError('%d: must be in %d..%d' % (n,n0,n1))        
    else:
      m = PAT.match(i.strip())
      if not m:
        raise ValueError('%r: illegal number' % i)
      b = n0
      if m.group(1):
        b = int(m.group(1))
      e = n1
      if m.group(2):
        e = int(m.group(2))
      if e < b:
        (b,e) = (e,b)
      if b < n0:
        raise ValueError('%d: must be in %d..%d' % (b,n0,n1))
      if n1 < e:
        raise ValueError('%d: must be in %d..%d' % (e,n0,n1))
      r.extend(range(b,e+1))
  return r

# read MPEG frames
BITRATE1 = [0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320]
BITRATE2 = [0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160]
SAMPLERATE1 = [44100, 48000, 32000]
SAMPLERATE2 = [22050, 24000, 16000]
SAMPLERATE25 = [11025, 12000, 8000]
def read_frames(frames, fp, verbose=False):
  while 1:
    x = fp.read(4)
    if not x:
      break
    elif x.startswith('TAG'):
      # TAG - ignored
      data = x[3]+fp.read(128-4)
      continue
    elif x.startswith('ID3'):
      # ID3 - ignored
      while 1:
        c = fp.read(1)
        if c == '' or c == '\xff': break
      x = c+fp.read(3)
    h = unpack('>L', x)[0]
    assert (h & 0xffe00000) == 0xffe00000, '!Frame Sync: %r' % x
    version = (h & 0x00180000) >> 19
    assert version != 1
    assert (h & 0x00060000) == 0x00020000, '!Layer3'
    protected = not (h & 0x00010000)
    b = (h & 0xf000) >> 12
    assert b != 0 and b != 15, '!Bitrate'
    if version == 3:                      # V1
      bitrate = BITRATE1[b]
    else:                                 # V2 or V2.5
      bitrate = BITRATE2[b]
    s = (h & 0x0c00) >> 10
    assert s != 3, '!SampleRate'
    if version == 3:                      # V1
      samplerate = SAMPLERATE1[s]
    elif version == 2:                    # V2
      samplerate = SAMPLERATE2[s]
    elif version == 0:                    # V2.5
      samplerate = SAMPLERATE25[s]
    nsamples = 1152
    if samplerate <= 24000:
      nsamples = 576
    pad = (h & 0x0200) >> 9
    channel = (h & 0xc0) >> 6
    joint = (h & 0x30) >> 4
    copyright = bool(h & 8)
    original = bool(h & 4)
    emphasis = h & 3
    if version == 3:
      framesize = 144000 * bitrate / samplerate + pad
    else:
      framesize = 72000 * bitrate / samplerate + pad
    data = x+fp.read(framesize-4)
    if verbose:
      print >>stderr, 'Frame%d: bitrate=%dk, samplerate=%d, framesize=%d' % \
            (len(frames), bitrate, samplerate, framesize)
    frames.append(data)
  return


# main
if __name__ == "__main__":
  import sys, getopt
  def usage():
    print "usage: edit_mp3.py [-v] expr [file ...] > new.mp3"
    sys.exit(2)
  try:
    (opts, args) = getopt.getopt(sys.argv[1:], "v")
  except getopt.GetoptError:
    usage()
  verbose = False
  for (k, v) in opts:
    if k == "-v": verbose = True
  if len(args) < 2: usage()
  expr = args[0]
  #
  allframes = []
  for fname in args[1:]:
    fp = file(fname)
    read_frames(allframes, fp, verbose)
    fp.close()
  if expr:
    for f in range2list(expr, 0, len(allframes)-1):
      sys.stdout.write(allframes[f])
  else:
    print >>stderr, 'Total frames:', len(allframes)
