#!/usr/bin/env python
##
##  edit_vnc2swf.py 0.3 - vnc2swf editor/viewer.
##    by Yusuke Shinyama <yusuke at cs dot nyu dot edu>
##
##  Required: Python 2.3 or above, Pygame-1.6 or above (recommended)
##   (you can also use Python Imaging Library 1.1.4 instead of Pygame.)
##
##  This program allows you to edit a .swf (Macromedia Flash) file
##  generated by vnc2swf application. Also this program does
##  some optimization to reduce the movie size a little bit.
##
##  NOTICE: This program only supports a .swf file generated
##    by vnc2swf. The behaviour where other swf files are given
##    is not defined. Generally you cannot manipulate other swf
##    files in the way this program works.
##
##  TODO: decent error handling, separate code.
##


import sys, zlib, StringIO
from struct import pack, unpack
stderr = sys.stderr
lowerbound = max
upperbound = min


try:
  # pygame 1.6 or newer.
  import pygame
  pygame.init()
  def imgsize(img):
    return img.get_size()
  def create_image(w, h):
    return pygame.Surface((w, h))
  def create_image_from_string(w, h, data):
    return pygame.image.fromstring(data, (w, h), 'RGB')
  def vflip_image(img):
    return pygame.transform.flip(img, 0, 1)
  def crop_image(img, (x0,y0,x1,y1)):
    (w,h) = (x1-x0),(y1-y0)
    tmp = pygame.Surface((w,h))
    tmp.blit(img, (0,0), (x0,y0,w,h))
    return tmp
  def paste_image(dest, src, (x0, y0)):
    return dest.blit(src, (x0, y0))
  def save_image(img, fname):
    if not fname.endswith('.bmp'):
      print >>stderr, 'Warning: this format not supported by pygame, raw rgb is used instead.'
    return pygame.image.save(img, fname)
  def convert_image_to_string(img):
    return pygame.image.tostring(img, 'RGB')
except ImportError:
  pygame = None
  try:
    # use PIL instead
    import Image
    def imgsize(img):
      return img.size
    def create_image(w, h):
      return Image.new('RGB', (w, h))
    def create_image_from_string(w, h, data):
      return Image.fromstring('RGB', (w, h), data)
    def vflip_image(img):
      return img.transpose(Image.FLIP_TOP_BOTTOM)
    def crop_image(img, (x0,y0,x1,y1)):
      return img.crop((x0, y0, x1, y1))
    def paste_image(dest, src, (x0, y0)):
      return dest.paste(src, (x0, y0))
    def save_image(img, fname):
      return img.save(fname)
    def convert_image_to_string(img):
      return img.tostring()
  except ImportError:
    print >>stderr, 'Either Pygame or Python Imaging Library is required.'  
  

##  MP3Reader
##
class MP3Reader:
  '''read MPEG frames'''
  def __init__(self):
    self.stereo = None
    self.bitrate = None
    self.samplerate = None
    self.frames = []
    self.totalsamples = 0L
    return

  def read(self, n):
    if self.length != None:
      if self.length <= 0:
        return ''
      self.length -= n
    return self.fp.read(n)
  
  BITRATE1 = [0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320]
  BITRATE2 = [0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160]
  SAMPLERATE1 = [44100, 48000, 32000]
  SAMPLERATE2 = [22050, 24000, 16000]
  SAMPLERATE25 = [11025, 12000, 8000]
  def readfile(self, fp, length=None):
    self.fp = fp
    self.length = length
    while 1:
      x = self.read(4)
      if not x: break
      if x.startswith('TAG'):
        # TAG - ignored
        data = x[3]+self.read(128-4)
        continue
      elif x.startswith('ID3'):
        # ID3 - ignored
        while 1:
          c = self.read(1)
          if c == '' or c == '\xff': break
        x = c+self.read(3)
      h = unpack('>L', x)[0]
      assert (h & 0xffe00000L) == 0xffe00000L, '!Frame Sync: %r' % x
      version = (h & 0x00180000L) >> 19
      assert version != 1
      assert (h & 0x00060000L) == 0x00020000L, '!Layer3'
      protected = not (h & 0x00010000L)
      b = (h & 0xf000) >> 12
      assert b != 0 and b != 15, '!Bitrate'
      if version == 3:                      # V1
        bitrate = self.BITRATE1[b]
      else:                                 # V2 or V2.5
        bitrate = self.BITRATE2[b]
      if self.bitrate != None and self.bitrate != bitrate:
        print >>stderr, 'warning: bitrate is changed.'
      else:
        self.bitrate = bitrate
      s = (h & 0x0c00) >> 10
      assert s != 3, '!Mp3'
      if version == 3:                      # V1
        samplerate = self.SAMPLERATE1[s]
      elif version == 2:                    # V2
        samplerate = self.SAMPLERATE2[s]
      elif version == 0:                    # V2.5
        samplerate = self.SAMPLERATE25[s]
      if self.samplerate != None and self.samplerate != samplerate:
        print >>stderr, 'warning: mp3_samplerate is changed.'
      else:
        self.samplerate = samplerate
      nsamples = 1152
      if samplerate <= 24000:
        nsamples = 576
      pad = (h & 0x0200) >> 9
      channel = (h & 0xc0) >> 6
      stereo = 1-(channel/2)
      if self.stereo != None and self.stereo != stereo:
        print >>stderr, 'warning: channel mode is changed.'
      else:
        self.stereo = stereo
      joint = (h & 0x30) >> 4
      copyright = bool(h & 8)
      original = bool(h & 4)
      emphasis = h & 3
      if version == 3:
        framesize = 144000 * bitrate / samplerate + pad
      else:
        framesize = 72000 * bitrate / samplerate + pad
      data = x+self.read(framesize-4)
      self.frames.append((nsamples, data))
      self.totalsamples += nsamples
    return (len(self.frames), self.totalsamples)


##  SWFParser
##
class SWFParser:
  """SWF Parser"""

  def __init__(self, debug=False):
    self.fp = None
    self.buff = 0
    self.bpos = 8
    self.debug = debug
    return

  # fixed bytes read
  
  def read(self, n):
    return self.fp.read(n)
  
  def readui8(self):
    return ord(self.fp.read(1))
  def readsi8(self):
    return unpack('<b', self.fp.read(1))[0]
  
  def readui16(self):
    return unpack('<H', self.fp.read(2))[0]
  def readub16(self):
    return unpack('>H', self.fp.read(2))[0]
  def readsi16(self):
    return unpack('<h', self.fp.read(2))[0]
  
  def readui32(self):
    return unpack('<L', self.fp.read(4))[0]

  def readrgb(self):
    return ( self.readui8(), self.readui8(), self.readui8() )
  def readrgba(self):
    return ( self.readui8(), self.readui8(), self.readui8(), self.readui8() )

  # fixed bits read

  def setbuff(self, bpos=8, buff=0):
    (self.bpos, self.buff) = (bpos, buff)
    return
  
  def readbits(self, bits, signed=False):
    if bits == 0: return 0
    bits0 = bits
    v = 0
    while 1:
      r = 8-self.bpos # the number of remaining bits we can get from the current buffer.
      if bits <= r:
        # |-----8-bits-----|
        # |-bpos-|-bits-|  |
        # |      |----r----|
        v = (v<<bits) | ((self.buff>>(r-bits)) & ((1<<bits)-1))
        self.bpos += bits
        break
      else:
        # |-----8-bits-----|
        # |-bpos-|---bits----...
        # |      |----r----|
        v = (v<<r) | (self.buff & ((1<<r)-1))
        bits -= r
        self.buff = ord(self.fp.read(1))
        self.bpos = 0
    if signed and (v>>(bits0-1)):
      v -= (1<<bits0)
    return v
  
  # variable length structure

  def readstring(self):
    s = []
    while 1:
      c = self.read(1)
      if c == '\x00': break
      s.append(c)
    return unicode(''.join(s), self.encoding)
  
  def readrect(self):
    '''(xmin, xmax, ymin, ymax) NOT width and height!'''
    x = ord(self.fp.read(1))
    bits = x>>3
    self.setbuff(5, x)
    return ( self.readbits(bits,1), self.readbits(bits,1), self.readbits(bits,1), self.readbits(bits,1) )

  def readmatrix(self):
    '''returns (scalex, scaley, rot0, rot1, transx, transy)'''
    self.setbuff()
    (scalex, scaley) = (None, None)
    if self.readbits(1):                # hasscale
      n = self.readbits(5)
      scalex = self.readbits(n,1)/65536.0
      scaley = self.readbits(n,1)/65536.0
    (rot0, rot1) = (None, None)
    if self.readbits(1):                # hasrotate
      n = self.readbits(5)
      rot0 = self.readbits(n,1)/65536.0
      rot1 = self.readbits(n,1)/65536.0
    (transx, transy) = (None, None)
    n = self.readbits(5)
    transx = self.readbits(n,1)
    transy = self.readbits(n,1)
    return (scalex, scaley, rot0, rot1, transx, transy)

  def readgradient(self, version):
    n = self.readui8()
    r = []
    for i in xrange(n):
      ratio = self.readui8()
      if version < 3:
        color = self.readrgb()
      else:
        color = self.readrgba()
      r.append((ratio, color))
    return r

  def read_style(self, version):
    '''
    fillstyles: list of (color, matrix, gradient, bitmapid, bitmapmatrix)
    linestyles: list of (width, color)
    '''
    # fillstylearray
    fillstyles = []
    nfills = self.readui8()
    if 2 <= version and nfills == 0xff:
      nfills = self.readui16()
    for i in xrange(nfills):
      t = self.readui8()
      (color, matrix, gradient, bitmapid, bitmapmatrix) = (None, None, None, None, None)
      if t == 0x00:
        if version == 3:
          color = self.readrgba()
        else:
          color = self.readrgb()
      elif t in (0x10, 0x12):
        matrix = self.readmatrix()
        gradient = self.readgradient(version)
      elif t in (0x40, 0x41, 0x42, 0x43):
        bitmapid = self.readui16()
        bitmapmatrix = self.readmatrix()
      fillstyles.append((color, matrix, gradient, bitmapid, bitmapmatrix))
    # linestylearray
    linestyles = []
    nlines = self.readui8()
    if 2 <= version and nlines == 0xff:
      nlines = self.readui16()
    for i in xrange(nlines):
      width = self.readui16()
      if version == 3:
        color = self.readrgba()
      else:
        color = self.readrgb()
      linestyles.append((width, color))
    return (fillstyles, linestyles)

  def read_shape(self, version, fillstyles=[], linestyles=[]):
    self.setbuff()
    nfillbits = self.readbits(4)
    nlinebits = self.readbits(4)
    r = []
    while 1:
      typeflag = self.readbits(1)
      if typeflag:
        # edge spec.
        straightflag = self.readbits(1)
        if straightflag:
          # StraightEdgeRecord
          n = self.readbits(4)+2
          if self.readbits(1):
            dx = self.readbits(n,1)
            dy = self.readbits(n,1)
            r.append( (1,(dx,dy)) )
          elif self.readbits(1):
            dy = self.readbits(n,1)
            r.append( (1,(0,dy)) )
          else:
            dx = self.readbits(n,1)
            r.append( (1,(dx,0)) )
        else:
          # CurveEdgeRecord
          n = self.readbits(4)+2
          cx = self.readbits(n,1)
          cy = self.readbits(n,1)
          ax = self.readbits(n,1)
          ay = self.readbits(n,1)
          r.append( (2,(cx,cy),(ax,ay)) )
      else:
        # style spec.
        flags = self.readbits(5)
        if flags == 0: break
        if flags & 1:
          n = self.readbits(5)
          x0 = self.readbits(n,1)
          y0 = self.readbits(n,1)
          r.append( (0,(x0,y0)) )
        if flags & 2:
          fillstyle0 = self.readbits(nfillbits)
        if flags & 4:
          fillstyle1 = self.readbits(nfillbits)
        if flags & 8:
          linestyle1 = self.readbits(nlinebits)
        if flags & 16:
          assert version
          (fillstyles, linestyles) = self.read_style(version)
          nfillbits = self.readbits(4)
          nlinebits = self.readbits(4)
    return r

  #

  def parse_header(self):
    (F,W,S,V) = self.read(4)
    assert W == 'W'
    assert S == 'S'
    self.swfversion = ord(V)
    if 6 <= self.swfversion:
      self.encoding = 'utf-8'
    self.totallen = self.readui32()
    if self.debug:
      print 'Header:', F,W,S,self.swfversion,self.totallen
    if F == 'C':
      # compressed
      x = zlib.decompress(self.fp.read())
      self.totallen -= 8
      assert len(x) == self.totallen, 'invalid tag'
      self.fp = StringIO.StringIO(x)
    self.rect = self.readrect()
    self.framerate = self.readui16()/256.0
    self.framecount = self.readui16()
    if self.debug:
      print 'Header:', self.rect, self.framerate, self.framecount
    return

  def parse_file(self, fname):
    print >>stderr, 'Reading source swf file: %s...' % fname
    self.fp = file(fname, 'rb')
    self.parse_header()
    while self.fp.tell() < self.totallen:
      self.parse_tag1()
    self.fp.close()
    return
  
  def parse_tag1(self):
    x = self.readui16()
    tag = x>>6
    if x & 63 == 63:
      length = self.readui32()
    else:
      length = x & 63
    pos0 = self.fp.tell()
    name = 'do_tag%d' % tag
    # branch to do_tag<N>
    if hasattr(self, name):
      getattr(self, name)(tag, length)
    else:
      self.do_unknown_tag(tag, length)
    self.fp.seek(pos0+length)
    return

  def do_unknown_tag(self, tag, length):
    if self.debug:
      print >>stderr, 'unknown tag: %d, length=%d' % (tag, length)
    return

  def do_tag0(self, tag, length):
    # end
    return


##  SWFWriter
##

# return the number of required bits for x.
def needbits1(x, signed=False):
  if x == 0:
    return 0
  if signed:
    n = 1
    if x < 0:
      x = -x-1
  else:
    n = 0
    assert 0 < x
  while 1:
    n += 1
    x >>= 1
    if x == 0: break
  return n

def needbits(args, signed=False):
  return max([ needbits1(x, signed) for x in args ])

# assert needbits1(0,0) == 0
# assert needbits1(0,1) == 0
# assert needbits1(1,0) == 1
# assert needbits1(1,1) == 2
# assert needbits1(2,0) == 2
# assert needbits1(-2,1) == 2
# assert needbits1(-3,1) == 3
# assert needbits1(127,0) == 7
# assert needbits1(127,1) == 8
# assert needbits1(128,0) == 8
# assert needbits1(-128,1) == 8
# assert needbits1(-129,1) == 9
# assert needbits1(-6380,1) == 14

class SWFWriter:
  """SWF Writer"""

  def __init__(self, outfile, swfversion, rect, framerate, compression):
    self.outfp = file(outfile, 'wb')
    self.swfversion = swfversion
    self.rect = rect
    self.framerate = framerate
    self.compression = compression
    self.fpstack = []
    self.bpos = 0
    self.buff = 0
    self.objid = 0
    if self.compression:
      self.fp = StringIO.StringIO()
      self.fp.write('CWS%c' % self.swfversion)
    else:
      self.fp = self.outfp
      self.fp.write('FWS%c' % self.swfversion)
    self.lenpos = self.fp.tell()
    self.writeui32(0) # dummy length
    self.writerect(rect)
    self.writeui16(int(framerate * 256))
    self.fcpos = self.fp.tell()
    self.writeui16(0) # dummy framecount
    return

  def newid(self):
    self.objid += 1
    assert self.objid < 65536, 'the number of objects exceeded the limit!'
    return self.objid

  def push(self):
    self.fpstack.append(self.fp)
    self.fp = StringIO.StringIO()
    return
  def pop(self):
    assert self.fpstack, 'empty fpstack'
    self.fp.seek(0)
    data = self.fp.read()
    self.fp = self.fpstack.pop()
    return data

  # fixed bytes write
  
  def write(self, *args):
    for x in args:
      self.fp.write(x)
    return

  def writeui8(self, *args):
    for x in args:
      self.fp.write(chr(x))
    return
  def writesi8(self, *args):
    for x in args:
      self.fp.write(pack('<b', x))
    return
  
  def writeui16(self, *args):
    for x in args:
      self.fp.write(pack('<H', x))
    return 
  def writeub16(self, *args):
    for x in args:
      self.fp.write(pack('>H', x))
    return 
  def writesi16(self, *args):
    for x in args:
      self.fp.write(pack('<h', x))
    return
  
  def writeui32(self, *args):
    for x in args:
      self.fp.write(pack('<L', x))
    return

  def writergb(self, (r,g,b)):
    self.writeui8(r,g,b)
    return
  def writergba(self, (r,g,b,a)):
    self.writeui8(r,g,b,a)
    return

  # fixed bits write
  def writebits(self, bits, x, signed=False):
    if signed and x < 0:
      x += (1<<bits)
    assert 0 <= x and x < (1<<bits)
    while 1:
      r = 8-self.bpos # the number of remaining bits we can write into the current buffer.
      if bits <= r:
        # |-----8-bits-----|
        # |-bpos-|-bits-|  |
        # |      |----r----|
        self.buff |= x << (r-bits)
        self.bpos += bits               # <= 8
        break
      else:
        # |-----8-bits-----|
        # |-bpos-|---bits----...
        # |      |----r----|
        self.fp.write(chr(self.buff | (x >> (bits-r)))) # r < bits
        self.buff = 0
        self.bpos = 0
        bits -= r                      # cut the upper r bits
        x &= (1<<bits)-1
    return
  
  def finishbits(self):
    if self.bpos:
      self.fp.write(chr(self.buff))
      self.buff = 0
      self.bpos = 0
    return
  
  # variable length structure
  
  def writestring(self, s):
    assert '\x00' not in s
    self.write(s)
    self.write('\x00')
    return

  def writerect(self, (a,b,c,d)):
    '''(xmin, xmax, ymin, ymax) NOT width and height!'''
    n = needbits((a,b,c,d), 1)
    self.writebits(5, n)
    self.writebits(n, a, 1)
    self.writebits(n, b, 1)
    self.writebits(n, c, 1)
    self.writebits(n, d, 1)
    self.finishbits()
    return

  def writematrix(self, (scalex,scaley, rot0,rot1, transx,transy)):
    if scalex != None:
      scalex = int(scalex*65536)
      scaley = int(scaley*65536)
      self.writebits(1, 1)
      n = needbits((scalex,scaley), 1)
      assert n < 32, 'too many bits needed'
      self.writebits(5, n)
      self.writebits(n, scalex, 1)
      self.writebits(n, scaley, 1)
    else:
      self.writebits(1, 0)
    if rot0 != None:
      self.writebits(1, 1)
      rot0 = int(rot0*65536)
      rot1 = int(rot1*65536)
      n = needbits((rot0,rot1), 1)
      assert n < 32, 'too many bits needed'
      self.writebits(5, n)
      self.writebits(n, rot0, 1)
      self.writebits(n, rot1, 1)
    else:
      self.writebits(1, 0)
    n = needbits((transx, transy), 1)
    assert n < 32, 'too many bits needed'
    self.writebits(5, n)
    self.writebits(n, transx, 1)
    self.writebits(n, transy, 1)
    self.finishbits()
    return

  def write_style(self, version, fillstyles, linestyles):
    # fillstyle
    if 254 < len(fillstyles):
      assert 2 <= version
      self.writeui8(0xff)
      self.writeui16(len(fillstyles))
    else:
      self.writeui8(len(fillstyles))
    for (t,color,matrix,gradientmatrix,bitmapid,bitmapmatrix) in fillstyles:
      self.writeui8(t)
      if color != None:
        assert t == 0x00
        if version == 3:
          self.writergba(color)
        else:
          self.writergb(color)
      elif matrix != None and gradient != None:
        assert t in (0x10, 0x12)
        self.writematrix(gradientmatrix)
        self.writegradient(gradient, version)
      elif bitmapid != None and bitmapmatrix != None:
        assert t in (0x40, 0x41, 0x42, 0x43)
        self.writeui16(bitmapid)
        self.writematrix(bitmapmatrix)
      else:
        assert 0, 'illegal arg: %d' % t
    # linestyle
    if 254 < len(linestyles):
      assert 2 <= version
      self.writeui8(0xff)
      self.writeui16(len(linestyles))
    else:
      self.writeui8(len(linestyles))
    for (width,color) in linestyles:
      self.writeui16(width)
      if version == 3:
        self.writergba(color)
      else:
        self.writergb(color)
    return

  # simple shape - up to only one fill/linestyle
  def write_shape(self, version, points, fillstyle=None, linestyle=None):
    assert 2 <= len(points)
    flags = 0 # moveto
    fillbits = 0
    if fillstyle != None:
      fillbits = needbits1(fillstyle)
      flags |= 2 # fillstyle1
    linebits = 0
    if linestyle != None:
      linebits = needbits1(linestyle)
      flags |= 4 # linestyle
    self.writebits(4, fillbits)
    self.writebits(4, linebits)
    # set the style and initial position
    (t,(x0,y0)) = points[0]
    assert t == 0
    self.writebits(1, 0) # typeflag:0
    self.writebits(4, flags)
    self.writebits(1, 1) # moveto:1
    n = needbits((x0,y0), 1)
    self.writebits(5, n)
    self.writebits(n, x0, 1)
    self.writebits(n, y0, 1)
    if fillstyle != None:
      self.writebits(fillbits, fillstyle)
    if linestyle != None:
      self.writebits(linebtis, linestyle)
    # put points
    for pt in points[1:]:
      t = pt[0]
      self.writebits(1, 1) # edgetype
      if t == 1:
        (t,(dx,dy)) = pt
        self.writebits(1, 1) # StraightEdgeRecord
        if dx == 0:
          n = lowerbound(needbits1(dy, 1), 2)
          self.writebits(4, n-2)
          self.writebits(1, 0) # write only
          self.writebits(1, 1) # dy
          self.writebits(n, dy, 1)
        elif dy == 0:
          n = lowerbound(needbits1(dx, 1), 2)
          self.writebits(4, n-2)
          self.writebits(1, 0) # write only
          self.writebits(1, 0) # dx
          self.writebits(n, dx, 1)
        else:
          n = lowerbound(needbits((dx,dy), 1), 2)
          self.writebits(4, n-2)
          self.writebits(1, 1) # write dx and dy
          self.writebits(n, dx, 1)
          self.writebits(n, dy, 1)
      elif t == 2:
        (t,(cx,cy),(ax,ay)) = pt
        n = lowerbound(needbits((cx,cy,ax,ay), 1), 2)
        self.writebits(4, n-2)
        self.writebits(n, cx, 1)
        self.writebits(n, cy, 1)
        self.writebits(n, ax, 1)
        self.writebits(n, ay, 1)
        (x0,y0) = (x2,y2)
      else:
        assert 0
    self.writebits(1, 0) # styletype
    self.writebits(5, 0) # EOS
    self.finishbits()
    return

  def start_tag(self):
    self.push()
    return
  def end_tag(self, tag, forcelong=False):
    data = self.pop()
    if 63 <= len(data) or forcelong:
      self.writeui16(tag << 6 | 63)
      self.writeui32(len(data))
    else:
      self.writeui16(tag << 6 | len(data))
    self.write(data)
    return

  def write_file(self, framecount):
    assert not self.fpstack
    self.fp.seek(0, 2) # seek to the last
    length = self.fp.tell()
    self.fp.seek(self.lenpos)
    self.writeui32(length) # adjust the length
    self.fp.seek(self.fcpos)
    self.writeui16(framecount) # adjust the framecount
    if self.compression:
      self.fp.seek(0)
      data = self.fp.read(8)
      self.outfp.write(data)
      data = self.fp.read()
      self.outfp.write(zlib.compress(data))
    self.outfp.close()
    return


################################################################


##  VNC2SWF_Parser
##
class VNC2SWF_Parser(SWFParser):

  def __init__(self, player, readmp3, debug=False):
    SWFParser.__init__(self, debug)
    self.player = player
    self.readmp3 = readmp3
    return

  def parse_header(self):
    SWFParser.parse_header(self)
    (x,width, y,height) = self.rect
    self.player.set_framerate(self.framerate)
    self.player.set_size((width, height))
    print >>stderr, 'Movie size: %dx%d, framerate=%dfps, frames=%d, duration=%.1fs.' % \
          (width/20, height/20, self.framerate, self.framecount, self.framecount/float(self.framerate))
    self.latseek = None
    self.image1_data = None
    self.image1_cid = None
    self.shape1_cid = None
    self.video1_cid = None
    self.images = []
    self.othertags = []
    return

  def do_tag1(self, tag, length):
    # ShowFrame
    if self.debug:
      print 'ShowFrame'
    self.player.reg_srcframe(self.images, self.othertags)
    self.images = []
    self.othertags = []
    return
  
  def do_unknown_tag(self, tag, length):
    data = self.read(length)
    self.othertags.append((tag, data))
    return
  
  def do_tag9(self, tag, length):
    # SetBackgroundColor
    bgcolor = self.readrgb()
    if self.debug:
      print 'BGColor:', bgcolor
    return
  
  def do_tag18(self, tag, length):
    # SoundStreamHead
    if not self.readmp3: return
    flags1 = self.readui8()
    flags2 = self.readui8()
    playrate = (flags1 & 0x0c) >> 2
    assert flags1 & 2 # playbacksoundsize
    playstereo = flags1 & 1
    compression = (flags2 & 0xf0) >> 4
    assert compression == 2 # must be mp3
    streamrate = (flags2 & 0x0c) >> 2
    assert streamrate != 0
    streamrate = [0,11025,22050,44100][streamrate]
    assert flags2 & 2 # streamsoundsize
    streamstereo = flags2 & 1
    avgsamplecount = self.readui16()
    self.latseek = self.readui16()
    self.player.set_mp3header(streamstereo, streamrate, self.latseek)
    if self.debug:
      print 'SoundStreamHeader', flags1, flags2, avgsamplecount, self.latseek
    return
  
  def do_tag19(self, tag, length):
    # SoundStreamBlock
    if not self.readmp3: return
    nsamples = self.readui16()
    seeksamples = self.readsi16()
    reader = MP3Reader()
    reader.readfile(self.fp, length-4)
    assert self.latseek == 0 or self.latseek == seeksamples
    self.latseek = 0
    if self.debug:
      print 'Block', nsamples, seeksamples, [x for (x,y) in reader.frames]
    self.player.reg_mp3frames(reader.frames)
    return
  
  def do_tag20(self, tag, length):
    # DefineBitsLossless
    cid = self.readui16()
    fmt = self.readui8()
    width = self.readui16()
    height = self.readui16()
    length -= 7
    tablesize = 0
    if fmt == 3:
      tablesize = self.readui8()+1
      length -= 1
    assert fmt == 5                     # RGB
    data = self.read(length)
    if self.debug:
      print 'DefineBitsLossless:', cid, fmt, width, height, len(data)
    self.image1_data = (width, height, data)
    self.image1_cid = cid
    return
  
  def do_tag32(self, tag, length):
    # DefineShape3
    sid = self.readui16()
    rect = self.readrect()
    (fillstyles, linestyles) = self.read_style(3)
    shape = self.read_shape(3, fillstyles, linestyles)
    assert self.image1_cid and fillstyles and fillstyles[0][3] == self.image1_cid
    if self.debug:
      print 'Shape', sid, rect, shape, fillstyles, linestyles
    self.shape1_cid = sid
    return

  def do_tag26(self, tag, length):
    # PlaceObject2
    flags = self.readui8()
    depth = self.readui16()
    (cid, matrix, ratio, name) = (None, None, None, None)
    if flags & 2:
      cid = self.readui16()
    matrix = None
    if flags & 4:
      matrix = self.readmatrix()
    assert not (flags & 8)
    if flags & 16:
      ratio = self.readui16()
    if flags & 32:
      name = self.readstring()
    assert not (flags & 64)
    assert not (flags & 128)
    #
    if self.debug:
      print 'Place', flags, depth, cid, matrix
    if not cid or cid == self.video1_cid:
      # ignore video frame
      return
    assert self.shape1_cid == cid
    assert matrix
    (scalex,scaley, rot0,rot1, transx,transy) = matrix
    self.images.append(((transx/20, transy/20), self.image1_data))
    self.image1_data = None
    self.image1_cid = None
    self.shape1_cid = None
    return
  
  def do_tag28(self, tag, length):
    # RemoveObject2
    depth = self.readui16()
    if self.debug:
      print 'RemoveObject', depth
    return

  def do_tag60(self, tag, length):
    # DefineVideoStream
    assert self.video1_cid == None, 'DefineVideoStream already appeared.'
    self.video1_cid = self.readui16()
    frames = self.readui16()
    width = self.readui16()
    height = self.readui16()
    flags = self.readui8() # ignore this.
    codec = self.readui8() # must be ScreenVideo
    assert codec == 3
    if self.debug:
      print 'DefineVideoStream', self.video1_cid, frames, width, height, flags, codec
    return
  
  def do_tag61(self, tag, length):
    # VideoFrame
    stream_id = self.readui16()
    assert self.video1_cid == stream_id, 'Video ID does not match.'
    framenum = self.readui16()
    self.setbuff()
    (frametype, codecid) = self.readbits(4), self.readbits(4)
    assert codecid == 3 # must be ScreenVideo
    (blockwidth, imagewidth) = self.readbits(4), self.readbits(12)
    (blockheight, imageheight) = self.readbits(4), self.readbits(12)
    blockwidth = (blockwidth+1)*16
    blockheight = (blockheight+1)*16
    if self.debug:
      print 'VideoFrame', framenum, frametype, ':',  blockwidth, imagewidth, blockheight, imageheight
    hblocks = (imagewidth+blockwidth-1)/blockwidth
    vblocks = (imageheight+blockheight-1)/blockheight
    for y in xrange(0, vblocks):
      for x in xrange(0, hblocks):
        length = self.readub16()
        if length:
          data = self.read(length)
          x0 = x*blockwidth
          y0 = imageheight-(y+1)*blockheight
          w = upperbound(blockwidth, imagewidth-x0)
          h = blockheight
          if y0 < 0:
            h += y0
            y0 = 0
          self.images.append( ((x0,y0), (w,h,data)) )
    return
  

##  SWF Screen Emulation
##
class SWFScreen:
  
  def __init__(self, x0, y0, w, h):
    (self.x0, self.y0, self.width, self.height) = (x0, y0, w, h)
    self.buf = create_image(w, h)
    return
  
  def __repr__(self):
    return '<SWFScreen: %dx%d at (%d, %d)>' % (self.width, self.height, self.x0, self.y0)

  def get_image(self, x, y, w, h):
    #assert 0 <= x and 0 <= y, (x,y)
    return crop_image(self.buf, (x, y, upperbound(x+w, self.width), upperbound(y+h, self.height)))
  
  def paint_image(self, data, x0, y0, w, h):
    x0 -= self.x0
    y0 -= self.y0
    if not (w and h and 0 < x0+w and x0 < self.width and 0 < y0+h and y0 < self.height):
      return False
    data = zlib.decompress(data)
    if len(data) == (w*h*4):
      # image defined by DefineBitsLossless
      data = ''.join([ data[i+1:i+4] for i in xrange(0, len(data), 4) ])
      img = create_image_from_string(w, h, data)
    elif len(data) == (w*h*3):
      # image defined by SCREENVIDEOPACKET
      data = ''.join([ data[i+2]+data[i+1]+data[i] for i in xrange(0, len(data), 3) ])
      img = vflip_image(create_image_from_string(w, h, data))
    else:
      assert 0, 'illegal image format: %s %s' % (len(data),(w,h))
    # sometime the pasted image doesn't fit my screenbuffer, but just let it out.
    paste_image(self.buf, img, (x0, y0))
    return True


class SWFBlockScreen(SWFScreen):
  
  def __init__(self, x0, y0, w, h, block_w, block_h):
    SWFScreen.__init__(self, x0, y0, w, h)
    self.block_w = block_w
    self.block_h = block_h
    self.hblocks = (w+self.block_w-1)/self.block_w
    self.vblocks = (h+self.block_h-1)/self.block_h
    return



##  SWFStorage
##
class SWFStorage:

  def __init__(self, clipping, debug):
    self.clipping = clipping
    self.debug = debug    
    # original movie data
    self.framerate = None # overridden by commandline option or preceding movies
    self.mp3sampleskip = None # overridden by commandline option or preceding movies
    self.width = None
    self.height = None
    self.srcframes = []
    self.srctags = []
    # original mp3 data
    self.streamrate = None # overridden by augmented mp3 or preceding movies
    self.streamstereo = None # overridden by augmented mp3 or preceding movies
    self.mp3frames = []
    return

  # set size in TWIP (pix*20)
  def set_size(self, (w0, h0)):
    w0 /= 20
    h0 /= 20
    if self.width != None and (self.width != w0 or self.height != h0):
      print >>stderr, 'Warning: movie size already specified: %dx%d' % (self.width, self.height)
      return
    (self.width, self.height) = (w0, h0)
    if not self.clipping:
      self.clipping = (0, 0, self.width, self.height)      
    return
      
  def set_framerate(self, framerate):
    if self.framerate != None and self.framerate != framerate:
      print >>stderr, 'Warning: movie framerate is overridden.'
      return      
    self.framerate = framerate
    return
  
  def set_mp3header(self, stereo, mp3_samplerate, mp3sampleskip):
    if (self.streamstereo != None and 
        (self.streamstereo != stereo or self.streamrate != mp3_samplerate)) or \
       (self.mp3sampleskip != None and self.mp3sampleskip != mp3sampleskip):
      print >>stderr, 'Warning: movie mp3 header is overridden.'
      return      
    self.streamstereo = stereo
    self.streamrate = mp3_samplerate
    self.mp3sampleskip = mp3sampleskip
    return

  # register
  def reg_srcframe(self, images, othertags):
    assert isinstance(images, list)
    self.srcframes.append(images)
    self.srctags.append(othertags)
    return
  
  def reg_mp3frames(self, mp3frames):
    self.mp3frames.extend(mp3frames)
    return

  def html_output(self, out, outfile):
    (w,h) = (self.screen.width, self.screen.height)
    out.write('<html><body>\n')
    out.write('<object classid="clsid:D27CDB6E-AE6D-11cf-96B8-444553540000"'
              ' width="%d" height="%d"\n' % (w,h))
    out.write(' codebase="http://download.macromedia.com/pub/shockwave/cabs/flash/swflash.cab#version=5,0,0,0">\n')
    out.write('<param name="movie" value="%s">\n' % outfile)
    out.write('<param name="play" value="true">\n')
    out.write('<param name="loop" value="true">\n')
    out.write('<param name="quality" value="high">\n')
    out.write('<embed src="%s" width="%d" height="%d"\n' % (outfile, w, h))
    out.write(' play="true" align="" loop="true" quality="high" type="application/x-shockwave-flash"\n')
    out.write(' pluginspage="http://www.macromedia.com/go/getflashplayer"></embed>\n')
    out.write('</object>\n')
    out.write('</body></html>\n')
    return



##  SWFVirtualPlayer
##
class SWFVirtualPlayer(SWFStorage):

  def __init__(self, outfile, frameskip, clipping, compression, debug):
    SWFStorage.__init__(self, clipping, debug)
    # parameters
    self.outfile = outfile
    self.frameskip = frameskip
    self.compression = compression
    # output movie working area
    self.writer = None
    self.got_mp3samples = None
    self.cur_frame = None
    self.cur_mp3frame = 0
    self.output_frames = 0
    self.frametodo = [] # skipped frame (if any)
    return

  # these parameters can be only set once:
  
  # mp3
  def seek_mp3frame(self, i):
    needsamples = int(self.streamrate / self.framerate)*i + self.mp3sampleskip
    self.cur_mp3frame = 0
    totalsamples = 0
    while self.cur_mp3frame < len(self.mp3frames):
      (samples,data) = self.mp3frames[self.cur_mp3frame]
      if needsamples <= totalsamples+samples: break
      totalsamples += samples
      self.cur_mp3frame += 1
    if self.debug:
      print >>stderr, 'seek_mp3frame:', self.cur_mp3frame
    return
  
  def write_mp3frame(self, i):
    needsamples = int(self.streamrate*(i+1) / self.framerate + self.mp3sampleskip)
    nsamples = 0
    mp3frames = []
    if needsamples <= self.got_mp3samples:
      #             |got_mp3samples
      #       |needsamples
      seeksamples = needsamples-self.got_mp3samples
    else:
      #       |got_mp3samples
      #           |s0       |needsamples
      #             or
      #               |got_mp3samples
      #           |s0       |needsamples
      s0 = int(self.streamrate*i / self.framerate + self.mp3sampleskip)
      seeksamples = s0 - self.got_mp3samples
      while self.cur_mp3frame < len(self.mp3frames):
        (samples,data) = self.mp3frames[self.cur_mp3frame]
        if needsamples < self.got_mp3samples+nsamples+samples: break
        nsamples += samples
        mp3frames.append(data)
        self.cur_mp3frame += 1
      if self.debug:
        print >>stderr, i, 'gotsample:', self.got_mp3samples, s0, needsamples, seeksamples
    # SoundStreamBlock
    self.writer.start_tag()
    self.writer.writeui16(nsamples)
    self.writer.writeui16(seeksamples)
    for data in mp3frames:
      self.writer.write(data)
    self.writer.end_tag(19)
    self.got_mp3samples += nsamples
    return

  def seek_frame(self, n):
    raise NotImplementedError
  def step_frame(self, n):
    raise NotImplementedError

  def write_frame1(self, n):
    # place objects
    if self.cur_frame != None and n == self.cur_frame+1:
      self.frametodo.append(n)
      if self.frameskip <= len(self.frametodo):
        self.step_frame(self.frametodo)
        self.frametodo = []
      else:
        self.cur_frame = n
        return False
    else:
      self.seek_frame(n)
      self.frametodo = []
      if self.mp3frames:
        self.seek_mp3frame(n)
    self.cur_frame = n
    # add other unknown tags
    for (tag, data) in self.srctags[self.cur_frame]:
      self.writer.start_tag()
      self.writer.write(data)
      self.writer.end_tag(tag)
    # insert sounds
    if self.mp3frames:
      self.write_mp3frame(self.output_frames)
    # ShowFrame
    self.writer.start_tag()
    self.writer.end_tag(1)
    self.output_frames += 1
    return True

  MP3_RATE = {11025:1, 22050:2, 44100:3}
  def start_movie(self):
    # init parameters
    self.got_mp3samples = self.mp3sampleskip
    # Write BGColor
    self.writer.start_tag()
    self.writer.writergb((255,255,255))
    self.writer.end_tag(9)
    if self.mp3frames:
      # write SoundStreamHeader
      assert self.streamstereo != None
      assert self.streamrate != None
      self.writer.start_tag()
      rate = self.MP3_RATE[self.streamrate]
      self.writer.writeui8(rate << 2 | 2 | self.streamstereo)
      self.writer.writeui8(rate << 2 | (2<<4) | 2 | self.streamstereo)
      self.writer.writeui16(int(self.streamrate / self.framerate))
      self.writer.writeui16(0)
      self.writer.end_tag(18)
    return

  def finish_movie(self):
    self.writer.start_tag()
    self.writer.end_tag(0)
    self.writer.write_file(self.output_frames)
    return


##  PygamePlayer
##
class PygamePlayer(SWFStorage):

  INTERVAL = 10
  def __init__(self,  clipping, debug):
    SWFStorage.__init__(self, clipping, debug)
    self.cur_frame = None
    self.ticks = 0
    return

  FONTSIZE = 24
  def set_size(self, (w0, h0)):
    SWFStorage.set_size(self, (w0, h0))
    (x,y,w,h) = self.clipping
    self.scrwidth = w
    self.scrheight = h
    self.scrbuf = SWFScreen(x, y, w, h)
    self.font = pygame.font.SysFont(pygame.font.get_default_font(), self.FONTSIZE)
    (fw1,fh1) = self.font.size('#0000  ')
    (fw2,fh2) = self.font.size('[>]  ')
    self.panel_x0 = 0
    self.panel_x1 = fw1
    self.panel_x2 = fw1+fw2
    self.panel_y0 = self.scrheight
    self.panel_y1 = self.scrheight + fh1/2
    self.panel_h = fh1
    self.panel_w = self.scrwidth-fw1-fw2-4
    self.slide_h = fh1/2
    self.slide_w = 8
    self.playing = True
    self.cur_frame = 0
    self.preserved = {}
    self.window = pygame.display.set_mode((w, self.scrheight+self.panel_h))
    return

  def step_frame(self):
    for ((x0,y0), (w,h,data)) in self.srcframes[self.cur_frame]:
      self.scrbuf.paint_image(data, x0, y0, w, h)
    self.cur_frame += 1
    return

  def seek_frame(self, frame):
    n = (frame/self.INTERVAL)*self.INTERVAL
    if n in self.preserved:
      self.scrbuf.buf.blit(self.preserved[n], (0,0))
      start_frame = n+1
    elif self.cur_frame == None or frame < self.cur_frame:
      start_frame = 0
    else:
      start_frame = self.cur_frame
    for i in xrange(start_frame, frame+1):
      for ((x0,y0), (w,h,data)) in self.srcframes[i]:
        self.scrbuf.paint_image(data, x0, y0, w, h)
    self.cur_frame = frame
    return

  def show_status(self):
    s = '#%04d' % self.cur_frame
    self.window.fill((0,0,0), (0, self.panel_y0, self.scrwidth, self.panel_h))
    self.window.blit(self.font.render(s, 0, (255,255,255)), (0, self.panel_y0))
    if self.playing:
      self.window.blit(self.font.render('[||]', 0, (255,0,0)), (self.panel_x1, self.panel_y0))
    else:
      self.window.blit(self.font.render('[>]', 0, (0,255,0)), (self.panel_x1, self.panel_y0))
    self.window.fill((255,255,255), (self.panel_x2, self.panel_y1, self.panel_w, 1))
    x = self.panel_x2 + self.panel_w*self.cur_frame/len(self.srcframes) - self.slide_w/2
    y = self.panel_y1 - self.slide_h/2
    self.window.fill((255,255,255), (x, y, self.slide_w, self.slide_h))
    return

  def goto_frame(self, n):
    if self.cur_frame != None and n == self.cur_frame+1:
      self.step_frame()
    else:
      self.seek_frame(n)
    if (n % self.INTERVAL == 0) and (n not in self.preserved):
      s = pygame.Surface((self.scrwidth, self.scrheight))
      s.blit(self.scrbuf.buf, (0,0))
      self.preserved[n] = s
    return

  def update(self):
    self.window.blit(self.scrbuf.buf, (0,0))
    self.show_status()
    pygame.display.update()
    if self.playing:
      ticks = pygame.time.get_ticks()
      d = lowerbound(0, int(1000/self.framerate) - (ticks-self.ticks))
      self.ticks = ticks
      pygame.time.wait(d)
    return

  def play(self, frames):
    n0 = -1
    n1 = 0
    drag = False
    loop = True
    while loop:
      if n0 != n1:
        self.goto_frame(frames[n1])
      n0 = n1
      self.update()
      if self.playing:
        events = pygame.event.get()
      else:
        events = [pygame.event.wait()]
      for e in events:
        if e.type in (pygame.MOUSEBUTTONDOWN, pygame.MOUSEMOTION):
          (x,y) = e.pos
          if (e.type == pygame.MOUSEBUTTONDOWN and 
              (y < self.panel_y0 or (self.panel_x1 <= x and x <= self.panel_x2))):
            # play/pause button
            self.playing = not self.playing
          elif (self.panel_y0 < y and self.panel_x2 < x and
                (e.type == pygame.MOUSEBUTTONDOWN or drag)):
            # slide bar dragging
            drag = True
            (x,y) = e.pos
            goal = (x-self.panel_x2)*len(self.srcframes)/self.panel_w
            goal = upperbound(lowerbound(goal, 0), len(self.srcframes))
            for i in xrange(len(frames)):
              if abs(frames[i]-goal) < abs(frames[n1]-goal):
                n1 = i
        elif e.type == pygame.MOUSEBUTTONUP:
          drag = False
        elif e.type == pygame.KEYDOWN:
          if e.key in (13, 32): # space or enter
            self.playing = not self.playing
          elif e.key in (113, 27): # 'q'uit, esc
            loop = False
          elif e.key == 275: # right
            n1 = upperbound(n1+1, len(frames)-1)
            self.playing = False
          elif e.key == 276: # left
            n1 = lowerbound(n1-1, 0)
            self.playing = False
          else:
            print 'Unknown key:', e
      if self.playing:
        n1 += 1
        if len(frames) <= n1:
          self.playing = False
          n1 = 0
    return
  


##  ImageConverter
##
class ImageConverter(SWFStorage):
    
  # this is dirty.
  # I should add another class between SWFVirtualPlayer and this.
  def __init__(self, basename, ext, clipping, debug):
    SWFStorage.__init__(self, clipping, debug)
    self.basename = basename
    self.ext = ext
    return

  def set_size(self, (w0, h0)):
    SWFStorage.set_size(self, (w0, h0))
    (x,y,w,h) = self.clipping
    self.screen = SWFScreen(x, y, w, h)
    self.cur_frame = None
    return
  
  def dump_frame1(self, frame):
    outfile = '%s-%04d.%s' % (self.basename, frame, self.ext)
    print >>stderr, 'Writing frame #%04d into %s...' % (frame, outfile)
    if self.cur_frame == None or frame < self.cur_frame:
      start_frame = 0
    else:
      start_frame = self.cur_frame
    for i in xrange(start_frame, frame+1):
      for ((x0,y0), (w,h,data)) in self.srcframes[i]:
        self.screen.paint_image(data, x0, y0, w, h)
    self.cur_frame = frame
    img = self.screen.get_image(0, 0, self.screen.width, self.screen.height)
    save_image(img, outfile)
    return


##  MP3Extractor
##
class MP3Extractor(SWFVirtualPlayer):
  
  def __init__(self, debug):
    SWFVirtualPlayer.__init__(self, '', 0, None, None, debug)
    return

  def dump_stream(self, fname):
    print >>stderr, 'Writing mp3 stream into %s...' % fname
    fp = file(fname, 'wb')
    for (sample,data) in self.mp3frames:
      fp.write(data)
    fp.close()
    return

  
##  ShapeConverter
##
class SWFObjectRef:
  
  def __init__(self, depth, count):
    self.depth = depth
    self.count = count
    return
  def __repr__(self):
    return '(%d,%d)' % (self.depth, self.count)

class SWFShapeScreen(SWFBlockScreen):
  BLOCKSIZE = 4
  
  def __init__(self, x0, y0, w, h):
    SWFBlockScreen.__init__(self, x0, y0, w, h, self.BLOCKSIZE, self.BLOCKSIZE)
    self.map = None
    return
  
  def remove_all(self):
    obj = SWFObjectRef(1, self.hblocks*self.vblocks)
    self.map = [ [obj]*self.hblocks for i in xrange(self.vblocks) ]
    return

  def paint_object(self, depth, x0, y0, w, h, removed):
    assert depth != 0
    x0 -= self.x0
    y0 -= self.y0
    #assert w and h and 0 < x0+w and x0 < self.width and 0 < y0+h and y0 < self.height
    x1 = upperbound((x0+w-1)/self.block_w+1, self.hblocks)
    y1 = upperbound((y0+h-1)/self.block_h+1, self.vblocks)
    #print x0,y0,w,h,self.block_w,self.block_h,',',x0/self.block_w,x1,y0/self.block_h,y1
    x0 = lowerbound(x0/self.block_w, 0)
    y0 = lowerbound(y0/self.block_h, 0)
    obj1 = SWFObjectRef(depth, (x1-x0)*(y1-y0))
    for line in self.map[y0:y1]:
      for x in xrange(x0, x1):
        obj0 = line[x]
        if obj0:
          obj0.count -= 1
          if obj0.count == 0:
            removed.append(obj0.depth)
        line[x] = obj1
    return (x0*self.block_w, y0*self.block_h, (x1-x0)*self.block_w, (y1-y0)*self.block_h)
  
  def is_painted(self, depth0, x0, y0, w, h):
    '''Check the area is already painted with the object depth >=depth0 ?'''
    assert depth0 != 0
    x0 -= self.x0
    y0 -= self.y0
    if x0+w <= 0 or self.width <= x0 or y0+h <= 0 or self.height <= y0:
      return True
    x1 = upperbound((x0+w)/self.block_w+1, self.hblocks)
    y1 = upperbound((y0+h)/self.block_h+1, self.vblocks)
    x0 = lowerbound(x0/self.block_w, 0)
    y0 = lowerbound(y0/self.block_h, 0)
    for line in self.map[y0:y1]:
      for obj0 in line[x0:x1]:
        if not obj0 or obj0.depth < depth0:
          return False
    return True
  
class ShapeConverter(SWFVirtualPlayer):

  def __init__(self, outfile, frameskip, clipping, compression, debug):
    SWFVirtualPlayer.__init__(self, outfile, frameskip, clipping, compression, debug)
    self.cur_objects = []
    return
  
  def set_size(self, (w0, h0)):
    SWFVirtualPlayer.set_size(self, (w0, h0))
    (x,y,w,h) = self.clipping
    self.screen = SWFShapeScreen(x, y, w, h)
    return

  def start_movie(self):
    # SWF5
    (x,y,w,h) = self.clipping
    self.writer = SWFWriter(self.outfile, 5, (0,w*20,0,h*20), self.framerate, self.compression)
    SWFVirtualPlayer.start_movie(self)
    return
  
  # add shape object
  def add_object(self, img, depth, x, y):
    (w,h) = imgsize(img)
    data = convert_image_to_string(img)
    data = zlib.compress(''.join([ '\x00'+data[i:i+3] for i in xrange(0,len(data),3) ]))
    if self.debug:
      print >>stderr, 'add_object:', depth, (x,y,w,h), len(data)
    self.cur_objects.append(depth)
    # DefineBitsLossless
    self.writer.start_tag()
    image_id = self.writer.newid()
    self.writer.writeui16(image_id)
    self.writer.writeui8(5)             # fmt=5
    self.writer.writeui16(w)
    self.writer.writeui16(h)
    self.writer.write(data)
    self.writer.end_tag(20, True) # forcelong because of flashplayer's bug
    # DefineShape3
    self.writer.start_tag()
    shape_id = self.writer.newid()
    self.writer.writeui16(shape_id)
    self.writer.writerect((20, w*20+20, 20, h*20+20))
    self.writer.write_style(3, [(0x43,None,None,None,image_id,(20,20,None,None,0,0))], [])
    self.writer.write_shape(3, [(0,(20,20)),(1,(w*20,0)),(1,(0,h*20)),(1,(-w*20,0)),(1,(0,-h*20))], fillstyle=1)
    self.writer.end_tag(32)
    # PlaceObject2
    self.writer.start_tag()
    self.writer.writeui8(2|4)
    self.writer.writeui16(depth)
    self.writer.writeui16(shape_id)
    self.writer.writematrix((None, None, None, None, x*20, y*20))
    self.writer.end_tag(26)
    return

  # remove shape object
  # if you leave objects on the screen, it gets verry slow.
  def remove_object(self, depth):
    if self.debug:
      print >>stderr, 'remove_object:', depth
    # RemoveObject2
    self.cur_objects.remove(depth)
    self.writer.start_tag()
    self.writer.writeui16(depth)
    self.writer.end_tag(28)
    return

  # step one frame
  def step_frame(self, frames):
    if self.debug:
      print >>stderr, 'step_frame:', frames
    assert self.cur_objects
    depth1 = max(self.cur_objects)+1
    depth0 = depth1
    addimages0 = []
    removeobjs = []
    for i in frames:
      for ((x0,y0), (w,h,data)) in self.srcframes[i]:
        if not self.screen.paint_image(data, x0, y0, w, h):
          continue
        # do not attempt to create another shape object if
        # its entire area is already taken by other objects which are
        # going to be created.
        if not self.screen.is_painted(depth0, x0, y0, w, h):
          (x0,y0,w,h) = self.screen.paint_object(depth1, x0, y0, w, h, removeobjs)
          addimages0.append( (depth1, x0, y0, w, h) )
          depth1 += 1
    addimages1 = []
    for (depth,x0,y0,w,h) in addimages0:
      if depth in removeobjs:
        removeobjs.remove(depth)
      else:
        addimages1.append((depth,x0,y0,w,h))
    # Remove completely overriden objects.
    for depth in removeobjs:
      self.remove_object(depth)
    for (depth,x0,y0,w,h) in addimages1:
      # Image & Shape & Place tags.
      self.add_object(self.screen.get_image(x0, y0, w, h), depth, x0, y0)
    return

  # seek to random frame
  def seek_frame(self, frame):
    if self.cur_frame == None or frame < self.cur_frame:
      start_frame = 0
    else:
      start_frame = self.cur_frame
    for depth in self.cur_objects[:]:
      self.remove_object(depth)
    assert self.cur_objects == []
    if self.debug:
      print >>stderr, 'seek_frame:', frame
    for i in xrange(start_frame, frame+1):
      for ((x0,y0), (w,h,data)) in self.srcframes[i]:
        self.screen.paint_image(data, x0, y0, w, h)
    self.add_object(self.screen.get_image(0, 0, self.screen.width, self.screen.height), 1, 0, 0)
    self.screen.remove_all()
    return


##  VideoConverter
##
class SWFVideoScreen(SWFBlockScreen):
  
  def __init__(self, x0, y0, w, h, block_w, block_h):
    SWFBlockScreen.__init__(self, x0, y0, w, h, block_w, block_h)
    self.block_changed = None
    self.block_image = None
    return

  def init_blocks(self):
    self.block_changed = [ [True]*self.hblocks for i in xrange(self.vblocks) ]
    self.block_image = [ [None]*self.hblocks for i in xrange(self.vblocks) ]
    return
  
  def get_block_change(self, x, y, forcechange):
    '''get change of block (x,y)'''
    if not forcechange and not self.block_changed[y][x]:
      return ''
    x0 = x*self.block_w
    y0 = self.height-(y+1)*self.block_h
    # if the block is partial, the player also expects a partial image.
    w = upperbound(self.block_w, self.width-x0)
    h = self.block_h
    if y0 < 0:
      h += y0
      y0 = 0
    # for some reason y-axis is filpped in VideoPacket.
    # so we flip it in advance so that it can go back correctly...
    img = vflip_image(self.get_image(x0, y0, w, h))
    data = convert_image_to_string(img)
    if not forcechange and self.block_image[y][x] == data:
      return ''
    self.block_changed[y][x] = False
    self.block_image[y][x] = data
    data = ''.join([ data[i+2]+data[i+1]+data[i] for i in xrange(0, len(data), 3) ])
    return data
  
  def paint_block(self, x0, y0, w, h):
    x0 -= self.x0
    y0 -= self.y0
    #assert w and h and 0 < x0+w and x0 < self.width and 0 < y0+h and y0 < self.height
    x1 = upperbound((x0+w-1)/self.block_w+1, self.hblocks)
    y1 = upperbound((self.height-y0)/self.block_h+1, self.vblocks)
    x0 = lowerbound(x0/self.block_w, 0)
    y0 = lowerbound((self.height-(y0+h-1))/self.block_h, 0)
    for line in self.block_changed[y0:y1]:
      for x in xrange(x0, x1):
        line[x] = True
    return
  
class VideoConverter(SWFVirtualPlayer):
  
  def __init__(self, outfile, frameskip, clipping, video_blocksize, compression, debug):
    SWFVirtualPlayer.__init__(self, outfile, frameskip, clipping, compression, debug)
    self.video_blocksize = video_blocksize
    self.video_object = None
    return
  
  def set_size(self, (w0, h0)):
    SWFVirtualPlayer.set_size(self, (w0, h0))
    (x,y,w,h) = self.clipping
    self.screen = SWFVideoScreen(x, y, w, h, self.video_blocksize, self.video_blocksize)
    return

  def add_videoframe(self, keyframe):
    #
    r = []
    exist = False
    for y in xrange(self.screen.vblocks):
      for x in xrange(self.screen.hblocks):
        data = self.screen.get_block_change(x, y, keyframe)
        r.append(data)
        if data:
          exist = True
    if not exist: return
    # write VideoFrame tag
    self.writer.start_tag()
    self.writer.writeui16(self.video_object) # video char
    self.writer.writeui16(self.output_frames) # frame num
    # SCREENVIDEOPACKET
    if keyframe:
      self.writer.writebits(4, 1)
    else:
      self.writer.writebits(4, 2)
    self.writer.writebits(4, 3) # screenvideo codec
    self.writer.writebits(4, self.screen.block_w/16-1)
    self.writer.writebits(12, self.screen.width)
    self.writer.writebits(4, self.screen.block_h/16-1)
    self.writer.writebits(12, self.screen.height)
    self.writer.finishbits()
    #
    for data in r:
      if data:
        data = zlib.compress(''.join(data))
      self.writer.writeub16(len(data))
      self.writer.write(data)
    self.writer.end_tag(61)
    # PlaceObject2
    # For some reason we need to set the RATIO to the current frame number every time.
    # This was not documented!
    self.writer.start_tag()
    self.writer.writeui8(17)
    self.writer.writeui16(self.video_object)
    self.writer.writeui16(self.output_frames)
    self.writer.end_tag(26)
    return

  def step_frame(self, frames):
    for i in frames:
      for ((x0,y0), (w,h,data)) in self.srcframes[i]:
        if self.screen.paint_image(data, x0, y0, w, h):
          self.screen.paint_block(x0, y0, w, h)
    self.add_videoframe(False)
    return

  def seek_frame(self, frame):
    if self.cur_frame == None or frame < self.cur_frame:
      start_frame = 0
    else:
      start_frame = self.cur_frame
    for i in xrange(start_frame, frame+1):
      for ((x0,y0), (w,h,data)) in self.srcframes[i]:
        self.screen.paint_image(data, x0, y0, w, h)
    self.screen.init_blocks()
    self.add_videoframe(True)
    return

  def start_movie(self):
    # SWF7
    (x,y,w,h) = self.clipping
    self.writer = SWFWriter(self.outfile, 7, (0,w*20,0,h*20), self.framerate, self.compression)
    SWFVirtualPlayer.start_movie(self)
    #
    self.video_object = self.writer.newid()
    # write DefineVideoStream
    assert not self.writer.fpstack
    pos0 = self.writer.fp.tell()
    self.writer.start_tag()
    self.writer.writeui16(self.video_object) # video char
    self.mangle_pos = pos0 + 4
    self.writer.writeui16(0)            # XXX: we need to change this later...so dirty
    self.writer.writeui16(self.screen.width)
    self.writer.writeui16(self.screen.height)
    self.writer.writeui8(0)             # smoothing off
    self.writer.writeui8(3)             # SCREENVIDEO
    self.writer.end_tag(60)
    # PlaceObject2
    self.writer.start_tag()
    self.writer.writeui8(2|4)
    self.writer.writeui16(1)            # depth
    self.writer.writeui16(self.video_object) # video char
    self.writer.writematrix((None, None, None, None, 0, 0))
    self.writer.end_tag(26)
    return
  
  def finish_movie(self):
    assert not self.writer.fpstack
    self.writer.fp.seek(self.mangle_pos) # mangle this
    self.writer.writeui16(self.output_frames) # set the number of frames into DefineVideoStream tag.
    self.writer.fp.seek(0, 2) # go back
    SWFVirtualPlayer.finish_movie(self)
    return


# range2list: converts strings like "1,5-8" to [1,5,6,7,8].
import re
PAT_RANGE = re.compile(r'^([0-9]*)-([0-9]*)$')
def range2list(s, n0, n1):
  r = []
  for i in s.split(','):
    i = i.strip()
    if not i: continue
    if i.isdigit():
      n = int(i)
      if n0 <= n and n <= n1:
        r.append(n)
      else:
        raise ValueError('%d: must be in %d..%d' % (n,n0,n1))        
    else:
      m = PAT_RANGE.match(i.strip())
      if not m:
        raise ValueError('%r: illegal number' % i)
      b = n0
      if m.group(1):
        b = int(m.group(1))
      e = n1
      if m.group(2):
        e = int(m.group(2))
      if e < b:
        (b,e) = (e,b)
      if b < n0:
        raise ValueError('%d: must be in %d..%d' % (b,n0,n1))
      if n1 < e:
        raise ValueError('%d: must be in %d..%d' % (e,n0,n1))
      r.extend(xrange(b,e+1))
  return r


# main
def main(argv):
  import getopt
  def usage():
    print >>stderr, '''usage: edit_vnc2swf.py  (version 0.2)
    [-d] [-c] [-H] [-V] [-f frames] [-a mp3file] [-r framerate]
    [-S mp3sampleskip] [-F frameskip] [-C WxH+X+Y] [-B block]
    -o outfile.{swf|png|bmp|mp3} swf1 swf2 ...

    Specify one output filename from the following:
      *.swf: generate a reorganized and/or augmented movie.
      *.png|*.bmp: save snapshots of given frames as "X-nnn.png"
      *.mp3: extract an MP3 audio stream from a movie.
      
    -d: debug mode.
    -c: compression.
    -H: generate an HTML file to frame the movie.
    -V: generate a movie in ScreenVideo format. (Flash version7 or above only)
    -f frames: frames to extract. e.g. 1-2,100-300,310,500-
    -a filename: attach MP3 file.
    -r framerate: override framerate.
    -B blocksize: blocksize of ScreenVideo format (must be a multiple of 16)
    -S N: skip the first N samples of the sound when the movie starts.
    -F N: multiply the movie speed N times.
    -C WxH+X+Y: crop a specific area of the movie.
    '''
    sys.exit(2)
  try:
    (opts, args) = getopt.getopt(argv, 'dr:o:VcHa:S:C:B:f:F:')
  except getopt.GetoptError:
    usage()
  #
  debug = 0
  outfile = ''
  mp3files = []
  framerange = '-'
  framerate = None
  mp3sampleskip = None
  frameskip = 1
  clipping = None
  videomode = False
  video_blocksize = 32
  compression = False
  outhtml = False
  for (k, v) in opts:
    if k == '-d': debug += 1
    elif k == '-r': framerate = float(v)
    elif k == '-o': outfile = v
    elif k == '-a': mp3files.append(v)
    elif k == '-S': mp3sampleskip = int(v)
    elif k == '-F': frameskip = int(v)
    elif k == '-C':
      m = re.match(r'^(\d+)x(\d+)\+(\d+)\+(\d+)$', v)
      if not m: usage()
      x = map(int, m.groups())
      clipping = (x[2],x[3], x[0],x[1])
    elif k == '-B':
      video_blocksize = int(v)
      assert 0 < video_blocksize and video_blocksize <= 256 and video_blocksize % 16 == 0
    elif k == '-V': videomode = True
    elif k == '-H': outhtml = True
    elif k == '-c': compression = True
    elif k == '-f': framerange = v
  if not args:
    print >>stderr, 'Specify at least one input movie.'
    usage()
    
  # play
  if not outfile:
    player = PygamePlayer(clipping, debug)
    for fname in args:
      VNC2SWF_Parser(player, not mp3files, debug).parse_file(fname)
    print >>stderr, 'Playing...'
    player.play(range2list(framerange, 0, len(player.srcframes)-1))

  # output .swf
  elif outfile.endswith('.swf'):
    # initialize
    if videomode:
      player = VideoConverter(outfile, frameskip, clipping, video_blocksize, compression, debug)
    else:
      player = ShapeConverter(outfile, frameskip, clipping, compression, debug)
    # read augmented mp3 file(s).
    for fname in mp3files:
      print >>stderr, 'Reading augmented mp3 file: %s...' % fname
      reader = MP3Reader()
      fp = file(fname, 'rb')
      (frames, samples) = reader.readfile(fp)
      if frames:
        print >>stderr, 'MP3 Info: frames=%d, duration=%.1fs, bitrate=%d' % \
              (frames, samples/reader.samplerate, reader.bitrate)
      fp.close()
      player.reg_mp3frames(reader.frames)
      if mp3sampleskip == None:
        mp3sampleskip = 0
      player.set_mp3header(reader.stereo, reader.samplerate, mp3sampleskip)
    # read the original movies.
    if framerate != None:
      player.set_framerate(framerate)
    for fname in args:
      VNC2SWF_Parser(player, not mp3files, debug).parse_file(fname)
    # play them.
    player.start_movie()
    stderr.write('Writing swf movie v%d to %s (%dx%d, framerate=%dfps)' % \
                 (player.writer.swfversion, outfile, player.width, player.height, player.framerate))
    for i in range2list(framerange, 0, len(player.srcframes)-1):
      if player.write_frame1(i):
        stderr.write('.')
        stderr.flush()
    stderr.write('%d frames total.\n' % player.output_frames)
    player.finish_movie()

  # output .png or .bmp
  elif outfile.endswith('.png') or outfile.endswith('.bmp'):
    # initialize
    player = ImageConverter(outfile[:-4], outfile[-3:], clipping, debug)
    # read the original movies.
    for fname in args:
      VNC2SWF_Parser(player, not mp3files, debug).parse_file(fname)
    # convert them.
    for i in range2list(framerange, 0, len(player.srcframes)-1):
      player.dump_frame1(i)

  # output .mp3
  elif outfile.endswith('.mp3'):
    # initialize
    player = MP3Extractor(debug)
    # read the original movies.
    for fname in args:
      VNC2SWF_Parser(player, not mp3files, debug).parse_file(fname)
    # extract.
    player.dump_stream(outfile)
    
  # other
  else:
    print >>stderr, 'unknown extension: %s' % outfile
    usage()

  # output html
  if outhtml:
    player.html_output(sys.stdout, outfile or args[0])

  sys.exit(0)


#
if __name__ == '__main__': main(sys.argv[1:])
